SystemOrganization addCategory: #'Ma contextual search-Domain'!

!MethodReference methodsFor: '*ma contextual search' stamp: 'cmm 12/8/2004 17:34'!
maContextKeywordsDo: aBlock

	{ self classSymbol.  self methodSymbol } do: aBlock! !

MaObject subclass: #MaAbstractContext
	instanceVariableNames: 'name'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma contextual search-Domain'!

!MaAbstractContext commentStamp: 'cmm 1/2/2005 17:58' prior: 0!
The purpose of this class is to provide a universal solution to the "Find by keyword" requirement.

My instances represent a human-relatable context in which a set of objects have relevance, and provide keyword-access to those objects.  It is intended to help users find objects.  Each of my subclasses provides its own particular implementation to accessing the objects in a context.  These implementations are stream-based and, therefore, very large contexts can be scanned.

A MauiCompositeContext can be used to aggregate contexts into single-keyword-searchable mega-context.  So, a "Squeak help-desk" context could be a composite-context of an in-image code-elements context, Swiki context, SqueakMap context and, finally, squeak-dev mailinglist archive context.  As searches within each context are kicked off in the background, results populate gradually so "research" can begin concurrently.

As each context scans, results are organized into MauiContextResults.  Currently, relevance-matching is based on a simple hierarchy of relevance, exact-match, case-insensitive match, begins-with match and contains match.   While these provide a lot of usability, this part needs to be extensible so we can support pluggable "sounds-like" or other matching algorithms.

Matching is performed by matching the keywords against the strings provided by any objects enumeration method, #maContextKeywordsDo:.!

!MaAbstractContext methodsFor: 'copying' stamp: 'cmm 12/8/2004 23:24'!
copyForIndependentSearch
	"Each instance maintains current-processing information, i.e., what am I doing right now and where am I?  Therefore, if you need to perform more than one search of a single context at a time, use this method to make a copy of me and use that one."

	self maMarked: 'dev'! !

!MaAbstractContext methodsFor: 'testing' stamp: 'cmm 12/11/2004 09:10'!
hasSubcontexts

	^ false! !

!MaAbstractContext methodsFor: 'initialize-release' stamp: 'cmm 1/2/2005 22:59'!
initializeKeywords
	self subclassResponsibility! !

!MaAbstractContext methodsFor: 'accessing' stamp: 'cmm 1/4/2005 21:43'!
keywords: anArray
	"Specify a an Array of keywords.  There must be some level of match on all keywords for an object to be matched."
	self initializeKeywords.
	anArray ifNotEmpty: [ self orKeywords: anArray ]! !

!MaAbstractContext methodsFor: 'private' stamp: 'cmm 12/10/2004 15:19'!
keywordsIn: aString
	^ aString substrings! !

!MaAbstractContext methodsFor: 'accessing' stamp: 'cmm 1/4/2005 21:44'!
keywordsString: aString
	"Specify one or keywords.  There must be at least a partial match on all keywords for an object to be matched."

	self keywords: (self keywordsIn: aString)! !

!MaAbstractContext methodsFor: 'testing' stamp: 'cmm 12/9/2004 15:04'!
knowsSize
	"Some contexts will be able to know their size, others won't.  If the size is known, you must implement #size and percent-complete can be known while scanning."
	^ false! !

!MaAbstractContext methodsFor: 'private' stamp: 'cmm 12/11/2004 09:08'!
maContextKeywordsDo: aBlock

	aBlock value: self name! !

!MaAbstractContext methodsFor: 'scanning' stamp: 'cmm 1/3/2004 17:37'!
mauiBasicApi

	^#(selectKeywords: results)! !

!MaAbstractContext methodsFor: 'accessing' stamp: 'cmm 6/21/2003 17:18'!
name

	^name! !

!MaAbstractContext methodsFor: 'accessing' stamp: 'cmm 6/21/2003 17:18'!
name: aString

	name _ aString! !

!MaAbstractContext methodsFor: 'accessing' stamp: 'cmm 1/2/2005 22:57'!
orKeywords: anArray

	self subclassResponsibility! !

!MaAbstractContext methodsFor: 'accessing' stamp: 'cmm 1/2/2005 22:56'!
orKeywordsString: keywordsString

	self orKeywords: (self keywordsIn: keywordsString)! !

!MaAbstractContext methodsFor: 'scanning' stamp: 'cmm 1/2/2005 22:25'!
refresh
	"Refresh the results with a new search."

	self subclassResponsibility! !

!MaAbstractContext methodsFor: 'scanning' stamp: 'cmm 1/2/2005 22:26'!
resetResults

	self subclassResponsibility! !

!MaAbstractContext methodsFor: 'accessing' stamp: 'cmm 12/8/2004 23:18'!
size
	"If your subclass answers true to #knowsSize, then you must override this method.  Otherwise, you don't have to."

	self subclassResponsibility! !

MaAbstractContext subclass: #MaCompositeContext
	instanceVariableNames: 'subcontexts'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma contextual search-Domain'!

!MaCompositeContext commentStamp: 'cmm 12/9/2004 15:22' prior: 0!
I represent the simple compositional element of no particular kind of Context.  Since MauiAbstractContext is an abstract class, you shouldn't use it as a concrete, outer context.!

!MaCompositeContext methodsFor: 'accessing' stamp: 'cmm 1/2/2005 23:09'!
addContext: aMauiContext

	subcontexts add: aMauiContext.
	aMauiContext keywords: self keywords! !

!MaCompositeContext methodsFor: 'initialize-release' stamp: 'cmm 1/5/2005 16:11'!
addDependent: aMauiCompositeContextMorph

	subcontexts do: [ : each | each addDependent: aMauiCompositeContextMorph ]! !

!MaCompositeContext methodsFor: 'testing' stamp: 'cmm 12/11/2004 09:11'!
hasSubcontexts

	^ subcontexts notEmpty! !

!MaCompositeContext methodsFor: 'initialize-release' stamp: 'cmm 12/11/2004 09:08'!
initialize
	super initialize.
	subcontexts _ OrderedCollection new! !

!MaCompositeContext methodsFor: 'initialize-release' stamp: 'cmm 1/2/2005 22:59'!
initializeKeywords
	subcontexts do: [ : each | each initializeKeywords ]! !

!MaCompositeContext methodsFor: 'accessing' stamp: 'cmm 1/2/2005 23:09'!
keywords

	^ subcontexts anyOne keywords! !

!MaCompositeContext methodsFor: 'accessing' stamp: 'cmm 12/11/2004 09:04'!
keywords: anArray

	subcontexts do: [ : each | each keywords: anArray ]! !

!MaCompositeContext methodsFor: 'testing' stamp: 'cmm 2/22/2004 19:22'!
knowsSize

	^ subcontexts allSatisfy: [ : each | each knowsSize ]! !

!MaCompositeContext methodsFor: 'accessing' stamp: 'cmm 1/2/2005 22:58'!
orKeywords: anArray

	subcontexts do: [ : each | each orKeywords: anArray ]! !

!MaCompositeContext methodsFor: 'private' stamp: 'cmm 1/2/2005 22:17'!
prepareForScan
	"A scan is about to begin.  Ensure you're ready to do it by initializing any resources and setting the position of the scan back to the beginning."

	subcontexts do: [ : each | each prepareForScan ]! !

!MaCompositeContext methodsFor: 'scan' stamp: 'cmm 12/11/2004 09:04'!
refresh

	subcontexts do: [ : each | each refresh ]! !

!MaCompositeContext methodsFor: 'initialize-release' stamp: 'cmm 1/2/2005 22:27'!
resetResults

	subcontexts do: [ : each | each resetResults ]! !

!MaCompositeContext methodsFor: 'accessing' stamp: 'cmm 1/2/2005 22:17'!
size

	^ self
		inject: 0
		into: [ : sum : each | sum + each size ]! !

!MaCompositeContext methodsFor: 'accessing' stamp: 'cmm 12/11/2004 09:06'!
subcontextNamed: aString

	^ subcontexts detect: [ : each | each name = aString ]! !

!MaCompositeContext methodsFor: 'accessing' stamp: 'cmm 1/4/2005 22:18'!
subcontexts
	^ subcontexts! !

MaAbstractContext subclass: #MaSearchContext
	instanceVariableNames: 'accessSelector keywords results preferences scanProcess'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma contextual search-Domain'!

MaSearchContext subclass: #MaContextForNonStreamable
	instanceVariableNames: 'queue queueingProcess doneQueueing'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma contextual search-Domain'!

!MaContextForNonStreamable commentStamp: 'cmm 12/9/2004 14:24' prior: 0!
This is an abstract class that allows contexts for objects that cannot implement #next, as required by MaAbstractContext.

Instead, my subclasses do their enumeration in a separate process and populate my 'queue' which is then accessed with my #next.

So my subclasses, instead of implementing #next, they simply do their enumeration and call my  #queue: method to add the objects to be scanned.!

MaContextForNonStreamable subclass: #MaAnyObjectContext
	instanceVariableNames: 'object traversalStrategy'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma contextual search-Domain'!

!MaAnyObjectContext commentStamp: 'cmm 12/8/2004 23:01' prior: 0!
I can search the graph of any object for objects matching a keyword.  I am useful for "default" searching and for developers who want to find objects quickly.

I have a MaObjectGraphTraversalStrategy that you may customize.!

!MaAnyObjectContext class methodsFor: 'create' stamp: 'cmm 12/9/2004 15:12'!
object: anObject
	"Create using the default traversal strategy."
	^ self new
		object: anObject ;
		yourself! !

!MaAnyObjectContext class methodsFor: 'create' stamp: 'cmm 12/9/2004 15:12'!
object: anObject traversalStrategy: aMaObjectGraphTraversalStrategy
	"Create specifying your own, customized aMaObjectGraphTraversalStrategy."
	^ (self object: anObject)
		traversalStrategy: aMaObjectGraphTraversalStrategy ;
		yourself! !

!MaAnyObjectContext methodsFor: 'initialize-release' stamp: 'cmm 1/2/2005 17:24'!
initialize
	super initialize.
	traversalStrategy _ 
		MaObjectGraphTraversalStrategy
			traverseWhen:
				[ : path : parent : index |
				path last class isPointers and: [ (self isGlobal: path last) not ] ] ;
			valueWhen: [ : path : parent : index | true ]! !

!MaAnyObjectContext methodsFor: 'testing' stamp: 'cmm 1/2/2005 17:24'!
isGlobal: anObject

	^ { World. ActiveHand. Transcript. Processor. Compiler. Sensor }! !

!MaAnyObjectContext methodsFor: 'accessing' stamp: 'cmm 12/9/2004 15:06'!
object

	^ object! !

!MaAnyObjectContext methodsFor: 'accessing' stamp: 'cmm 12/9/2004 15:06'!
object: anObject

	object _ anObject! !

!MaAnyObjectContext methodsFor: 'overriding' stamp: 'cmm 12/9/2004 15:13'!
queueObjectsToScan
	object
		maGraphDo:
			[ : path : parent : index |
			self queue: path last.
			self shouldStopScanning ifTrue: [ ^ self ] ]
		using: traversalStrategy! !

!MaAnyObjectContext methodsFor: 'accessing' stamp: 'cmm 12/8/2004 23:52'!
traversalStrategy

	^ traversalStrategy! !

!MaAnyObjectContext methodsFor: 'accessing' stamp: 'cmm 12/8/2004 23:52'!
traversalStrategy: aMaObjectGraphTraversalStrategy

	traversalStrategy _ aMaObjectGraphTraversalStrategy! !

MaContextForNonStreamable subclass: #MaCodeElementContext
	instanceVariableNames: 'scanSelector'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma contextual search-Domain'!

!MaCodeElementContext commentStamp: 'cmm 12/8/2004 15:42' prior: 0!
I represent a context for objects in the image.!

!MaCodeElementContext class methodsFor: 'as yet unclassified' stamp: 'cmm 1/3/2005 10:22'!
forClasses

	^ self new 
		name: 'Classes in this image' ;
		scanSelector: #scanClasses ;
		yourself! !

!MaCodeElementContext class methodsFor: 'as yet unclassified' stamp: 'cmm 1/3/2005 10:22'!
forImage

	^ MaCompositeContext new
		name: 'Code-elements in this image (packages, classes and methods)' ;
"		addContext: self forPackages ;  "
		addContext: self forClasses ;
		addContext: self forMethods ;
		yourself! !

!MaCodeElementContext class methodsFor: 'as yet unclassified' stamp: 'cmm 1/3/2005 10:21'!
forMethods

	^ self new 
		name: 'Methods in this image' ;
		scanSelector: #scanMethods;
		yourself! !

!MaCodeElementContext class methodsFor: 'as yet unclassified' stamp: 'cmm 8/7/2003 13:47'!
forPackages

	self maMarked: 'dev'! !

!MaCodeElementContext class methodsFor: 'as yet unclassified' stamp: 'cmm 1/3/2005 10:19'!
forSourceCode
	"By default, we only look for method names when looking for methods.  This adds each methods sourceCode as an enumerator."

	^ self forMethods
		accessSelector: #sourceCode ;
		yourself! !

!MaCodeElementContext methodsFor: 'overriding' stamp: 'cmm 12/9/2004 16:23'!
queueObjectsToScan

	self perform: scanSelector! !

!MaCodeElementContext methodsFor: 'initializing' stamp: 'cmm 12/9/2004 16:25'!
scanClasses

	SystemNavigation new allClasses do:
		[ : each |
		self queue: each.
		self shouldStopScanning ifTrue: [ ^ self ] ]! !

!MaCodeElementContext methodsFor: 'initializing' stamp: 'cmm 1/2/2005 23:31'!
scanMethods

	SystemNavigation new maAllMethodReferencesDo:
		[ : each |
		self queue: each.
		self shouldStopScanning ifTrue: [ ^ self ] ]! !

!MaCodeElementContext methodsFor: 'initializing' stamp: 'cmm 12/9/2004 15:21'!
scanSelector: aSymbol
	"The scanSelector is the selector of the method used to scan my objects."

	scanSelector _ aSymbol! !

!MaContextForNonStreamable methodsFor: 'overriding' stamp: 'cmm 12/9/2004 16:38'!
atEnd
	^ doneQueueing and: [ queue isEmpty ]! !

!MaContextForNonStreamable methodsFor: 'initialize-release' stamp: 'cmm 12/9/2004 16:37'!
initialize
	super initialize.
	queue _ SharedQueue new.
	doneQueueing _ false! !

!MaContextForNonStreamable methodsFor: 'overriding' stamp: 'cmm 12/9/2004 14:39'!
next
	^ queue next! !

!MaContextForNonStreamable methodsFor: 'overriding' stamp: 'cmm 12/9/2004 16:37'!
prepareForScan
	queue flush.
	doneQueueing _ false.
	queueingProcess _
		[ self queueObjectsToScan.
		doneQueueing _ true ] fork! !

!MaContextForNonStreamable methodsFor: 'private' stamp: 'cmm 12/9/2004 14:25'!
queue: anObject
	queue nextPut: anObject! !

!MaContextForNonStreamable methodsFor: 'subclass responsibility' stamp: 'cmm 12/9/2004 14:54'!
queueObjectsToScan
	"Subclasses, override this method and call #queue: to queue the objects you want scanned.  Also, be sure to check #shouldStopScanning very frequently to determine whether to abort your queuing."
	self subclassResponsibility! !

!MaContextForNonStreamable methodsFor: 'private' stamp: 'cmm 12/9/2004 15:01'!
stopScanning
	super stopScanning.
	"If you have correctly implemented your queueObjectsToScan method, then your queuingProcess will abort from the above."
	queueingProcess ifNotNil:
		[ [ queueingProcess isTerminated ] whileFalse: [ (Delay forMilliseconds: 100) wait ] ]! !

MaContextForNonStreamable subclass: #MaContextResultsContext
	instanceVariableNames: 'source resultsByRelevance'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma contextual search-Domain'!

!MaContextResultsContext commentStamp: 'cmm 12/9/2004 16:55' prior: 0!
I represent the results of the scan performed by some MaContext.  I provide categories for the results by a "relevance factor."  1 is the most relevant, higher numbers are less relevant.  The relevance factor is also, quite simply, the index into my resultsByRelevance.!

!MaContextResultsContext class methodsFor: 'creation' stamp: 'cmm 12/9/2004 22:34'!
source: aMauiContext

	^ self new
		source: aMauiContext;
		yourself! !

!MaContextResultsContext methodsFor: 'private' stamp: 'cmm 6/23/2003 17:52'!
atRelevance: relevanceIndex add: anObject

	(resultsByRelevance at: relevanceIndex) add: anObject! !

!MaContextResultsContext methodsFor: 'accessing' stamp: 'cmm 12/9/2004 22:35'!
atRelevanceNamed: aString

	^ resultsByRelevance at: (source indexOfRelevanceNamed: aString)! !

!MaContextResultsContext methodsFor: 'enumeration' stamp: 'cmm 12/10/2004 15:40'!
collect: aBlock 
	"Evaluate aBlock with each of the receiver's elements as the argument.  
	Collect the resulting values into a collection like the receiver. Answer  
	the new collection."

	| newCollection |
	newCollection _ self species new.
	self do: [:each | newCollection add: (aBlock value: each)].
	^ newCollection! !

!MaContextResultsContext methodsFor: 'enumeration' stamp: 'cmm 11/23/2003 17:52'!
do: aBlock

	resultsByRelevance do: [ : eachCollection | eachCollection do: aBlock ]! !

!MaContextResultsContext methodsFor: 'testing' stamp: 'cmm 12/11/2004 12:45'!
includes: anObject
	^ resultsByRelevance anySatisfy: [ : each | each includes: anObject ]! !

!MaContextResultsContext methodsFor: 'testing' stamp: 'cmm 12/10/2004 15:11'!
isEmpty
	^ self size = 0! !

!MaContextResultsContext methodsFor: 'overriding' stamp: 'cmm 12/9/2004 22:20'!
knowsSize
	^ true! !

!MaContextResultsContext methodsFor: 'enumeration' stamp: 'cmm 12/10/2004 15:46'!
maDo: aBlock while: conditionBlock

	conditionBlock value ifFalse: [ ^ self ].
	self do:
		[ : each |
		aBlock value: each.
		conditionBlock value ifFalse: [ ^ self ] ]! !

!MaContextResultsContext methodsFor: 'accessing' stamp: 'cmm 1/2/2005 20:48'!
maNewSearchContext 

	^ self class source: self! !

!MaContextResultsContext methodsFor: 'accessing' stamp: 'cmm 12/9/2004 17:10'!
mauiBasicApi

	^#(size asSet)! !

!MaContextResultsContext methodsFor: 'accessing' stamp: 'cmm 12/10/2004 15:10'!
mauiName

	| msg |
	msg _ MaTextMessage new.
	msg append: (self class name first isVowel ifTrue: ['an '] ifFalse: ['a ']).
		self size = 0 ifTrue: [ msg append: 'empty ' ].
		msg append: self class name.
		self isEmpty ifFalse:
			[ | classCounts |
			msg append: ' of '.
			classCounts _ (self collect: [ : each | each isBehavior ifTrue: [ Class ] ifFalse: [ each class ] ]) asBag sortedCounts.
			classCounts size > 1 ifTrue: [ msg append: 'mostly ' ].
			msg 
				appendProperlyPluralized: classCounts first value name
				for: classCounts first key ].
	^msg messageText! !

!MaContextResultsContext methodsFor: 'printing' stamp: 'cmm 12/10/2004 15:51'!
printOn: aStream

	super printOn: aStream.
	self do:
		[ : each | 
		aStream cr; maPrint: each ]! !

!MaContextResultsContext methodsFor: 'overriding' stamp: 'cmm 12/9/2004 22:53'!
queueObjectsToScan
	self do:
		[ : each |
		self queue: each.
		self shouldStopScanning ifTrue: [ ^ self ] ]! !

!MaContextResultsContext methodsFor: 'scanning' stamp: 'cmm 12/10/2004 13:48'!
refresh
	"Refreshing me from my resultsByRelevance makes no sense since those  won't change.  Instead, we tell my source context to refresh me, remembering that if my sourceContext is also a ResultsContext, it will likewise ask its source context, and onward until we reach the originating context with the store of objects.
	Therefore, this is query by example.  The user, having performed a series of narrowing searches, can re-execute all those searches with this single message send."

	source refresh! !

!MaContextResultsContext methodsFor: 'initialize-release' stamp: 'cmm 12/9/2004 23:21'!
resetResultsByRelevance
	resultsByRelevance _ (Array new: source relevanceNames size) collect: [ : each | OrderedCollection new ]! !

!MaContextResultsContext methodsFor: 'private' stamp: 'cmm 8/7/2003 13:38'!
resultsByRelevance

	^resultsByRelevance ! !

!MaContextResultsContext methodsFor: 'accessing' stamp: 'cmm 12/9/2004 17:09'!
size

	^ resultsByRelevance
		inject: 0
		into: [ : sum : each | sum + each size ]! !

!MaContextResultsContext methodsFor: 'initialize-release' stamp: 'cmm 12/9/2004 23:20'!
source: aMauiContext

	source _ aMauiContext.
	self resetResultsByRelevance! !

!MaContextResultsContext methodsFor: 'private' stamp: 'cmm 12/10/2004 15:41'!
species
	^ OrderedCollection! !

MaSearchContext subclass: #MaFilesystemContext
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma contextual search-Domain'!

!MaSearchContext methodsFor: 'accessing' stamp: 'cmm 1/2/2005 22:36'!
accessSelector

	^ accessSelector! !

!MaSearchContext methodsFor: 'accessing' stamp: 'cmm 1/2/2005 22:36'!
accessSelector: aSymbol
	"The second-level of filtering available after selectionKeys have been applied.  aSymbol will be performed to determine the object to match the selectionKeys."

	accessSelector _ aSymbol! !

!MaSearchContext methodsFor: 'subclass behavior' stamp: 'cmm 1/2/2005 22:16'!
atEnd

	self subclassResponsibility! !

!MaSearchContext methodsFor: 'private comparing' stamp: 'cmm 1/2/2005 22:22'!
bestScoreFor: compareObject
	"Answer an Integer that represents the best match among my keywords.  Note that for each Array of anded keywords, *all* of that Array must match with a minimum score, otherwise none of the Array is a match and we move to the next Array."

	| answer |
	answer _ self noMatchScore.
	compareObject maContextKeywordsDo:
		[ : eachKeyword |  | score |
		score _ self calculateScoreFor: eachKeyword.
		score ifNotNil: [ answer _ answer min: score ] ].
	^ answer = self noMatchScore ifTrue: [ nil ] ifFalse: [ answer ]! !

!MaSearchContext methodsFor: 'private comparing' stamp: 'cmm 1/2/2005 22:22'!
calculateScoreFor: compareString
	"Enumerate my sets of keywords to find the lowest score of the set."
	^ keywords
		inject: self noMatchScore
		into:
			[ : min : eachArray |  | setScore |
			setScore _
				self
					calculateSet: eachArray
					scoreFor: compareString.
			setScore min: min ]! !

!MaSearchContext methodsFor: 'private comparing' stamp: 'cmm 1/2/2005 22:23'!
calculateSet: keywordsArray scoreFor: compareString
	"Every word in keywordsArray is considered AND'ed, therefore must minimally match, otherwise the entire set is not a match."

	^ keywordsArray
		inject: 99
		into:
			[ : score : each |  | selectionScore |
			selectionScore _
				self
	
				scoreFor: compareString
	
				keyword: each.
			selectionScore ifNil: [ ^ self noMatchScore ].
			score min: selectionScore ]! !

!MaSearchContext methodsFor: 'private comparing' stamp: 'cmm 1/2/2005 22:34'!
getCompareObjectFor: anObject
	"If using the accessSelector feature, compare based on that, otherwise on anObject."
	^ accessSelector
		ifNil: [ anObject ]
		ifNotNil:
			[ (anObject respondsTo: accessSelector)
				ifTrue: [ anObject perform: accessSelector ]
				ifFalse: [ anObject ] ]! !

!MaSearchContext methodsFor: 'private' stamp: 'cmm 1/2/2005 22:28'!
indexOfRelevanceNamed: aString

	^ self relevanceNames 
		indexOf: aString
		ifAbsent:
			[ MauiUserError signal:
				(String streamContents:
					[ : stream |
					stream
						nextPutAll: aString ;
						nextPutAll: ' is not valid.  Must use any of '.
					self relevanceNames do:
						[ : each |
						stream
							nextPutAll: each ;
							space ] ]) ]! !

!MaSearchContext methodsFor: 'initialize-release' stamp: 'cmm 1/2/2005 22:52'!
initialize

	super initialize.
	self initializeKeywords.
	preferences _ MaContextPreferences new.
	scanProcess _ self newScanProcess! !

!MaSearchContext methodsFor: 'initialize-release' stamp: 'cmm 1/2/2005 22:53'!
initializeKeywords

	keywords _ OrderedCollection new! !

!MaSearchContext methodsFor: 'accessing' stamp: 'cmm 1/2/2005 23:07'!
keywords

	^ keywords! !

!MaSearchContext methodsFor: 'private' stamp: 'cmm 1/2/2005 22:30'!
newScanProcess

	^MaClientProcess doBlock: [ self scan ]! !

!MaSearchContext methodsFor: 'subclass behavior' stamp: 'cmm 1/2/2005 22:16'!
next
	"Answer the next object that should be evaluated against the keyword selection."

	"I keep track of my position (thus am meant to be used by a single Process at a time).  This method answers the next available element based on its current position.  If the subclass receiver has a way reducing the search-set and answering only the next that meets its selectionKey, it should.  Otherwise, they will be eliminated via serial searching."

	self subclassResponsibility! !

!MaSearchContext methodsFor: 'private comparing' stamp: 'cmm 1/2/2005 22:35'!
noMatchScore
	"A special number, higher than will ever be needed for any particular relevance-level, indicating no match."
	
	^ 99! !

!MaSearchContext methodsFor: 'accessing' stamp: 'cmm 1/2/2005 22:38'!
orKeywords: anArray

	keywords add: anArray! !

!MaSearchContext methodsFor: 'accessing' stamp: 'cmm 1/2/2005 22:25'!
preferences

	^preferences! !

!MaSearchContext methodsFor: 'accessing' stamp: 'cmm 1/2/2005 22:28'!
preferred: attributeSymbol

	^(preferences perform: attributeSymbol)
		ifNil: [ MaContextPreferences global perform: attributeSymbol ]! !

!MaSearchContext methodsFor: 'subclass behavior' stamp: 'cmm 1/2/2005 22:16'!
prepareForScan
	"A scan is about to begin.  Ensure you're ready to do it by initializing any resources and setting the position of the scan back to the beginning."

	"optional subclass responsibility"! !

!MaSearchContext methodsFor: 'private' stamp: 'cmm 1/2/2005 22:30'!
primScanUsing: aMaProgressNotification
	"Stream the next anInteger objects into aMauiContextResults according to my keywords and organized in aMauiContextResults by relevance."

	self prepareForScan.
self maMarked: 'dev'.
"If you ever want to be able to resume searching where you left off (say, if time ran out before any results were found), one cop-out implementation might be to simply suspend the scanProcess if the runningTime or results count conditions were met but not atEnd."
	[ self shouldStopScanning ]

		whileFalse:
			[ | score next |
			next _ self next.
			score _ self bestScoreFor: (self getCompareObjectFor: next).
			score ifNotNil:
				[ results
					atRelevance: score
					add: next.
				self changed ].
			aMaProgressNotification advance; signalIfTime ].
	scanProcess shouldStop ifTrue: [ scanProcess signalWaiterOnStop ]! !

!MaSearchContext methodsFor: 'scanning' stamp: 'cmm 1/2/2005 22:24'!
refresh

	self stopScanning.
	scanProcess _ self newScanProcess.
	scanProcess start! !

!MaSearchContext methodsFor: 'private' stamp: 'cmm 1/2/2005 22:30'!
relevanceNames
	"Answer an Array of Strings that name the kinds of relevance matching relates to each search result, found by having applied a String in some way.  This is the default, subclasses may override as necessary, but if they do, they will also need to override matchFactorBetween:and: if they override the behavior, as well as the names of the various kinds of matching."

	^#('exact' 'case-insensitive' 'starts with' 'contains')  "sounds-like?"! !

!MaSearchContext methodsFor: 'scanning' stamp: 'cmm 1/2/2005 22:26'!
resetResults

	results ifNil: [ results _ MaContextResultsContext source: self ].
	results resetResultsByRelevance.
	self changed! !

!MaSearchContext methodsFor: 'accessing' stamp: 'cmm 1/2/2005 22:26'!
results

	^results! !

!MaSearchContext methodsFor: 'private' stamp: 'cmm 1/2/2005 22:31'!
scan

	| progress description |
	progress _ MaProgressNotification new
		description: 'scanning for ', keywords printString;
		unitsVerbPhrase: 'objects scanned'.
	self knowsSize ifTrue: [ progress taskSize: self size ].
	progress signal.
	self
		resetResults ;
		primScanUsing: progress.
	description _
		scanProcess shouldStop
			ifTrue: [ 'stopped' ]
			ifFalse:
				[ self atEnd
					ifTrue: [ 'scanning complete' ]
					ifFalse:
						[ scanProcess runningTime > (self preferred: #maximumScanTime)
							ifTrue: [ 'timed out' ]
							ifFalse: [ 'max results found' ] ] ].
	progress
		description: description;
		signal! !

!MaSearchContext methodsFor: 'accessing' stamp: 'cmm 1/2/2005 22:26'!
scanProcess

	^scanProcess! !

!MaSearchContext methodsFor: 'private comparing' stamp: 'cmm 1/2/2005 22:35'!
scoreFor: objectString keyword: keywordString
	"Answer a magnitude suitable for sorting results according to best match."

	| nonCaseKeywordString nonCaseObjectString |
	keywordString = objectString ifTrue: [ ^ 1 ].
	nonCaseKeywordString _ keywordString asLowercase.
	nonCaseObjectString _ objectString asLowercase.
	nonCaseKeywordString = nonCaseObjectString ifTrue: [ ^ 2 ].
	(nonCaseObjectString beginsWith: nonCaseKeywordString) ifTrue: [ ^ 3 ].
	(nonCaseObjectString includesSubstring: nonCaseKeywordString caseSensitive: true) ifTrue: [ ^ 4 ].
	^ nil! !

!MaSearchContext methodsFor: 'testing' stamp: 'cmm 1/2/2005 22:29'!
shouldStopScanning
	^ scanProcess shouldStop
		or: [ self atEnd 
		or: [ scanProcess runningTime > (self preferred: #maximumScanTime)
		or: [ results size >= (self preferred: #maximumScanResults) ] ] ]! !

!MaSearchContext methodsFor: 'private' stamp: 'cmm 1/2/2005 22:32'!
stopScanning
	"If scan already in progress, terminate it."
	scanProcess ifNotNil:
		[ scanProcess isRunning ifTrue: [ scanProcess pleaseStop ].
		scanProcess terminateProcess ]! !

MaSearchContext subclass: #MaStreamContext
	instanceVariableNames: 'stream'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma contextual search-Domain'!

MaStreamContext subclass: #MaCollectionContext
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma contextual search-Domain'!

!MaCollectionContext class methodsFor: 'creation' stamp: 'cmm 6/21/2003 16:52'!
collection: aCollection 

	^self
		collection: aCollection
		accessSelector: nil! !

!MaCollectionContext class methodsFor: 'creation' stamp: 'cmm 11/23/2003 23:12'!
collection: aCollection accessSelector: aSymbol
	"aSymbol is the selector that will be performed on objects in aCollection to obtain the collection of compareStrings."

	^self 
		stream: aCollection readStream
		accessSelector: aSymbol! !

!MaCollectionContext methodsFor: 'initializing' stamp: 'cmm 7/6/2003 23:00'!
collection: aSequenceableCollection

	^self stream: aSequenceableCollection readStream! !

MaStreamContext subclass: #MaContextForComposite
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma contextual search-Domain'!

!MaContextForComposite commentStamp: 'cmm 12/9/2004 16:44' prior: 0!
I am a context used to search a composite domain structure.  To use, you specify the selector that accesses the "children" of each composite.!

!MaContextForComposite class methodsFor: 'private' stamp: 'cmm 7/20/2004 13:27'!
addChildrenOf: compositeObject to: streamsCollection using: selectorSymbol

	| children |
	children _ compositeObject perform: selectorSymbol.
	streamsCollection add: children readStream.
	children do:
		[ : each |
		self
			addChildrenOf: each
			to: streamsCollection
			using: selectorSymbol ]! !

!MaContextForComposite class methodsFor: 'create' stamp: 'cmm 7/20/2004 13:38'!
root: anObject childrenSelector: selectorSymbol

	| streams |
	streams _ OrderedCollection new.
	streams add: { anObject } readStream.
	self 
		addChildrenOf: anObject
		to: streams
		using: selectorSymbol.
	^ self stream: (MaCompositeReadStream streams: streams)! !

!MaStreamContext class methodsFor: 'create' stamp: 'cmm 6/21/2003 16:51'!
stream: aStream

	^self
		stream: aStream
		accessSelector: nil! !

!MaStreamContext class methodsFor: 'create' stamp: 'cmm 6/21/2003 16:53'!
stream: aStream accessSelector: aSymbol

	^self new
		stream: aStream;
		accessSelector: aSymbol;
		yourself! !

!MaStreamContext methodsFor: 'private' stamp: 'cmm 6/23/2003 16:24'!
atEnd

	^stream atEnd! !

!MaStreamContext methodsFor: 'private' stamp: 'cmm 6/21/2003 16:54'!
next

	^stream next! !

!MaStreamContext methodsFor: 'private' stamp: 'cmm 12/8/2004 23:45'!
prepareForScan

	stream reset! !

!MaStreamContext methodsFor: 'accessing' stamp: 'cmm 8/6/2003 00:35'!
size

	^stream size! !

!MaStreamContext methodsFor: 'private' stamp: 'cmm 6/21/2003 16:54'!
stream: aStream

	stream _ aStream! !

MaSearchContext subclass: #MagmaCollectionContext
	instanceVariableNames: 'reader position'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma contextual search-Domain'!

!MagmaCollectionContext class methodsFor: 'creation' stamp: 'cmm 6/21/2003 16:35'!
reader: aMagmaCollectionReader

	^self new
		reader: aMagmaCollectionReader;
		yourself ! !

!MagmaCollectionContext class methodsFor: 'creation' stamp: 'cmm 12/8/2004 23:54'!
reader: aMagmaCollectionReader accessSelector: aSymbol

	^ (self reader: aMagmaCollectionReader)
		accessSelector: aSymbol;
		yourself! !

!MagmaCollectionContext methodsFor: 'accessing' stamp: 'cmm 6/23/2003 16:25'!
atEnd

	^position >= reader size! !

!MagmaCollectionContext methodsFor: 'private' stamp: 'cmm 6/21/2003 16:38'!
incrementPosition

	^position _ position + 1! !

!MagmaCollectionContext methodsFor: 'private' stamp: 'cmm 6/21/2003 16:38'!
initialize

	position _ 0! !

!MagmaCollectionContext methodsFor: 'accessing' stamp: 'cmm 6/21/2003 16:38'!
next

	^reader at: self incrementPosition! !

!MagmaCollectionContext methodsFor: 'overrides' stamp: 'cmm 12/8/2004 23:46'!
prepareForScan

	| smallestResults |
	keywords do:
		[ : each |
		| eachReader |
		eachReader _ 
			self collection
				read: reader attribute
				at: each.
		smallestResults _ smallestResults
			ifNil: [ eachReader ]
			ifNotNil: [ { eachReader. smallestResults } detectMin: [ : challengingReader | challengingReader lastKnownSize ] ] ].
	reader _ smallestResults! !

!MagmaCollectionContext methodsFor: 'private' stamp: 'cmm 6/21/2003 16:37'!
reader: aMagmaCollectionReader

	reader _ aMagmaCollectionReader! !

!MagmaCollectionContext methodsFor: 'accessing' stamp: 'cmm 8/6/2003 00:35'!
size

	^reader size! !

MaObject subclass: #MaContextPreferences
	instanceVariableNames: 'maximumScanTime maximumScanResults'
	classVariableNames: 'Global'
	poolDictionaries: ''
	category: 'Ma contextual search-Domain'!

!MaContextPreferences class methodsFor: 'as yet unclassified' stamp: 'cmm 11/19/2003 00:44'!
global

	^Global! !

!MaContextPreferences class methodsFor: 'as yet unclassified' stamp: 'cmm 11/19/2003 00:45'!
initialize

	Global _ self new! !

!MaContextPreferences methodsFor: 'private' stamp: 'cmm 3/3/2004 13:21'!
initialize

	maximumScanTime _ 5 maMinutes.
	maximumScanResults _ 250! !

!MaContextPreferences methodsFor: 'api' stamp: 'cmm 1/19/2004 23:31'!
mauiBasicApi

	^#(maximumScanResults maximumScanTime)! !

!MaContextPreferences methodsFor: 'accessing' stamp: 'cmm 8/8/2003 13:52'!
maximumScanResults
	"The maximum number of results per subcontext that will be retrieved before moving to the next subcontext."

	^maximumScanResults! !

!MaContextPreferences methodsFor: 'accessing' stamp: 'cmm 6/24/2003 16:25'!
maximumScanResults: anInteger

	maximumScanResults _ anInteger! !

!MaContextPreferences methodsFor: 'accessing' stamp: 'cmm 8/8/2003 13:53'!
maximumScanTime
	"The maximum time this context will scan before passing control to the next context."

	^maximumScanTime! !

!MaContextPreferences methodsFor: 'accessing' stamp: 'cmm 6/24/2003 16:25'!
maximumScanTime: aMaDuration

	maximumScanTime _ aMaDuration! !

!Stream methodsFor: '*ma contextual search' stamp: 'cmm 1/2/2005 20:51'!
maNewSearchContext 

	^ MaStreamContext stream: self! !

!String methodsFor: '*ma contextual search' stamp: 'cmm 12/8/2004 17:33'!
maContextKeywordsDo: aBlock
	{ self } do: aBlock! !

Object subclass: #MaCompositeReadStream
	instanceVariableNames: 'streams currentInnerStream size'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma contextual search-Domain'!

!MaCompositeReadStream commentStamp: 'cmm 7/20/2004 12:37' prior: 0!
I allow multiple read-streams to be treated as one big stream.  Start by creating me with a Collection of ReadStreams and then you can say next, next, next and I will enumerate them all, in sequence.!

!MaCompositeReadStream class methodsFor: 'examples' stamp: 'cmm 7/20/2004 13:13'!
example1
	"self example1"
	| stream |
	stream _ self streams:
		{ (1 to: 10) asArray readStream.
		'abcxyz' readStream }.
	[ stream atEnd ] whileFalse: [ Transcript show: stream next printString ].
	stream reset.
	"again."
	[ stream atEnd ] whileFalse: [ Transcript show: stream next printString ]! !

!MaCompositeReadStream class methodsFor: 'create' stamp: 'cmm 7/20/2004 13:48'!
streams: aCollection

	^ self new
		streams: aCollection ;
		yourself! !

!MaCompositeReadStream methodsFor: 'access' stamp: 'cmm 7/20/2004 13:08'!
atEnd

	^ self currentInnerStream atEnd! !

!MaCompositeReadStream methodsFor: 'private' stamp: 'cmm 7/20/2004 13:07'!
currentInnerStream

	(currentInnerStream isNil or: [ currentInnerStream atEnd ])
		ifTrue:
			[ streams atEnd ifFalse: [ currentInnerStream _ streams next ] ].
	^ currentInnerStream! !

!MaCompositeReadStream methodsFor: 'access' stamp: 'cmm 7/20/2004 13:50'!
isEmpty

	^ self size = 0! !

!MaCompositeReadStream methodsFor: 'access' stamp: 'cmm 7/20/2004 13:07'!
next

	^ self currentInnerStream next! !

!MaCompositeReadStream methodsFor: 'access' stamp: 'cmm 7/20/2004 13:18'!
reset

	currentInnerStream _ nil.
	streams reset.
	[ streams atEnd ] whileFalse: [ streams next reset ].
	streams reset! !

!MaCompositeReadStream methodsFor: 'access' stamp: 'cmm 7/20/2004 13:50'!
size
	"Once created, ReadStreams cannot change in size so this should be accurate."

	^ size! !

!MaCompositeReadStream methodsFor: 'initialize' stamp: 'cmm 7/20/2004 13:49'!
streams: aCollectionOfStreams

	streams _ aCollectionOfStreams readStream.
	size _ aCollectionOfStreams
		inject: 0
		into: [ : sum : eachStream | sum + eachStream size ]! !

!Object methodsFor: '*ma contextual search' stamp: 'cmm 12/8/2004 16:43'!
maContextKeywordsDo: aBlock
	"Override this method in subclasses to allow my instances to be found by the MaContext framework.  Just calculate whatever keywords you want for me and tell aBlock to #value: each of those keywords.  This will cause evaluation against the user-entered keywords and weigh the relevance of each particular instance accordingly."

	{ self printString } do: aBlock! !

!Object methodsFor: '*ma contextual search' stamp: 'cmm 1/2/2005 18:16'!
maNewSearchContext

	^ MaAnyObjectContext object: self! !

!Collection methodsFor: '*ma contextual search' stamp: 'cmm 12/8/2004 17:32'!
maContextKeywordsDo: aBlock
	"Generally, a Collection won't be a single element of a larger container, but its certainly possible.  Nevertheless, we don't want to rely on my super implementation, (printString) which could easily truncate, causing inconsistent results."

	self do: [ : each | each maContextKeywordsDo: aBlock ]! !

!Collection methodsFor: '*ma contextual search' stamp: 'cmm 1/2/2005 20:50'!
maNewSearchContext 

	^ MaCollectionContext collection: self! !

!CompiledMethod methodsFor: '*ma contextual search' stamp: 'cmm 1/2/2005 23:29'!
maContextKeywordsDo: aBlock

	{ self selector } do: aBlock! !

!SystemDictionary methodsFor: '*ma contextual search' stamp: 'cmm 1/2/2005 20:46'!
maNewSearchContext 

	^ MaCodeElementContext forImage! !
MaContextPreferences initialize!
